@@ -42,6 +42,7 @@ use Module::Runtime qw/ use_module /;;
 use Weasel::FindExpanders qw/ expand_finder_pattern /;
 use Weasel::WidgetHandlers qw| best_match_handler_class |;
 
+
 =head1 ATTRIBUTES
 
 =over
@@ -63,7 +64,11 @@ has 'driver' => (is => 'ro',
 
 =item widget_groups
 
-Contains the list of widget groups to be
+Contains the list of widget groups to be used with the session, or
+uses all groups when undefined.
+
+Note: this functionality allows to load multiple groups into the running
+perl instance, while using different groups in various sessions.
 
 =cut
 
@@ -73,6 +78,9 @@ has 'widget_groups' => (is => 'rw');
 
 Holds the prefix that will be prepended to every URL passed
 to this API.
+The prefix can be an environment variable, e.g. ${VARIABLE}.
+It will be expanded and default to hppt://localhost:5000 if not defined.
+If it is not an environment variable, it will be used as is.
 
 =cut
 
@@ -98,6 +106,17 @@ sub _page_builder {
     return $class->new(session => $self);
 }
 
+=item log_hook
+
+Upon instantiation can be set to log consumer; a function of 3 arguments:
+ 1. the name of the event
+ 2. the text to be logged (or a coderef to be called without arguments returning such)
+
+=cut
+
+has 'log_hook' => (is => 'ro',
+                   isa => 'Maybe[CodeRef]');
+
 =item page_class
 
 Upon instantiation can be set to an alternative class name for the C<page>
@@ -135,6 +154,7 @@ has 'poll_delay' => (is => 'rw',
 
 =back
 
+
 =head1 METHODS
 
 
@@ -150,7 +170,8 @@ textarea elements and input elements of type text and password.
 sub clear {
     my ($self, $element) = @_;
 
-    $self->driver->clear($element->_id);
+    $self->_logged(sub { $self->driver->clear($element->_id); },
+                   'clear', 'clearing input element');
 }
 
 =item click([$element])
@@ -164,7 +185,11 @@ current mouse location.
 sub click {
     my ($self, $element) = @_;
 
-    $self->driver->click(($element) ? $element->_id : undef);
+    $self->_logged(
+        sub {
+            $self->driver->click(($element) ? $element->_id : undef);
+        },
+        'click', ($element) ? 'clicking element' : 'clicking window');
 }
 
 =item find($element, $locator [, scheme => $scheme] [, %locator_args])
@@ -179,11 +204,14 @@ sub find {
     my ($self, @args) = @_;
     my $rv;
 
-    $self->wait_for(
+    $self->_logged(
         sub {
-            my @rv = @{$self->find_all(@args)};
-            return $rv = shift @rv;
-        });
+            $self->wait_for(
+                sub {
+                    my @rv = @{$self->find_all(@args)};
+                    return $rv = shift @rv;
+                });
+        }, 'find', 'find ' . $args[1]);
 
     return $rv;
 }
@@ -202,13 +230,25 @@ sub find_all {
     my ($self, $element, $pattern, %args) = @_;
 
     my $expanded_pattern = expand_finder_pattern($pattern, \%args);
-    my @rv =
-        map { $self->_wrap_widget($_) }
-        $self->driver->find_all($element->_id,
-                                $expanded_pattern,
-                                $args{scheme});
-    print STDERR "found " . scalar(@rv) . " elements for $pattern " . (join(', ', %args)) . "\n";
-    print STDERR ' - ' . ref($_) . " (" . $_->tag_name . ")\n" for (@rv);
+    my @rv = $self->_logged(
+        sub {
+            return
+                map { $self->_wrap_widget($_) }
+                $self->driver->find_all($element->_id,
+                                        $expanded_pattern,
+                                        $args{scheme});
+        },
+        'find_all',
+        sub {
+            my ($rv) = @_;
+            return "found " . scalar(@$rv) . " elements for $pattern "
+                . (join(', ', %args)) . "\n"
+                . (join("\n",
+                        map { ' - ' . ref($_)
+                                  . ' (' . $_->tag_name . ")" } @$rv));
+        },
+        "pattern: $pattern");
+
     return wantarray ? @rv : \@rv;
 }
 
@@ -223,10 +263,16 @@ after prefixing with C<base_url>.
 sub get {
     my ($self, $url) = @_;
 
-    $url = $self->base_url . $url;
+    my $base = $self->base_url =~ /\$\{([a-zA-Z0-9_]+)\}/
+             ? $ENV{$1} // "http://localhost:5000"
+             : $self->base_url;
+    $url = $base . $url;
     ###TODO add logging warning of urls without protocol part
     # which might indicate empty 'base_url' where one is assumed to be set
-    $self->driver->get($url);
+    $self->_logged(
+        sub {
+            return $self->driver->get($url);
+        }, 'get', "loading URL: $url");
 }
 
 =item get_attribute($element, $attribute)
@@ -239,7 +285,10 @@ identified by C<$element>, or C<undef> if the attribute isn't defined.
 sub get_attribute {
     my ($self, $element, $attribute) = @_;
 
-    return $self->driver->get_attribute($element->_id, $attribute);
+    return $self->_logged(
+        sub {
+            return $self->driver->get_attribute($element->_id, $attribute);
+        }, 'get_attribute', "element attribute '$attribute'");
 }
 
 =item get_text($element)
@@ -251,7 +300,11 @@ Returns the 'innerHTML' of the element identified by C<$element>.
 sub get_text {
     my ($self, $element) = @_;
 
-    return $self->driver->get_text($element->_id);
+    return $self->_logged(
+        sub {
+            return $self->driver->get_text($element->_id);
+        },
+        'get_text', 'element text');
 }
 
 =item is_displayed($element)
@@ -265,7 +318,11 @@ the viewport for interaction.
 sub is_displayed {
     my ($self, $element) = @_;
 
-    return $self->driver->is_displayed($element->_id);
+    return $self->_logged(
+        sub {
+            return $self->driver->is_displayed($element->_id);
+        },
+        'is_displayed', 'query is_displayed');
 }
 
 =item screenshot($fh)
@@ -281,7 +338,10 @@ Note: this version assumes pictures of type PNG will be written;
 sub screenshot {
     my ($self, $fh) = @_;
 
-    $self->driver->screenshot($fh);
+    $self->_logged(
+        sub {
+            $self->driver->screenshot($fh);
+        }, 'screenshot', 'screenshot');
 }
 
 =item send_keys($element, @keys)
@@ -294,7 +354,11 @@ simulating keyboard input.
 sub send_keys {
     my ($self, $element, @keys) = @_;
 
-    $self->driver->send_keys($element->_id, @keys);
+    $self->_logged(
+        sub {
+            $self->driver->send_keys($element->_id, @keys);
+        },
+        'send_keys', 'sending keys: ' . join('', @keys));
 }
 
 =item tag_name($element)
@@ -306,7 +370,10 @@ Returns the tag name of the element identified by C<$element>.
 sub tag_name {
     my ($self, $element) = @_;
 
-    return $self->driver->tag_name($element->_id);
+    return $self->_logged(sub { return $self->driver->tag_name($element->_id) },
+         'tag_name',
+         sub { my $tag = shift; return "found tag with name $tag" },
+         'getting tag name');
 }
 
 =item wait_for($callback, [ retry_timeout => $number,] [poll_delay => $number])
@@ -322,12 +389,68 @@ session-global settings.
 sub wait_for {
     my ($self, $callback, %args) = @_;
 
-    $self->driver->wait_for($callback,
-                            retry_timeout => $self->retry_timeout,
-                            poll_delay => $self->poll_delay,
-                            %args);
+    $self->_logged(
+        sub {
+            $self->driver->wait_for($callback,
+                                    retry_timeout => $self->retry_timeout,
+                                    poll_delay => $self->poll_delay,
+                                    %args);
+        },
+        'wait_for', 'waiting for condition');
 }
 
+
+sub _appending_wrap {
+    my ($str) = @_;
+    return sub {
+        my $rv = shift;
+        if ($rv) {
+            return "$str ($rv)";
+        }
+        else {
+            return $str;
+        }
+    }
+}
+=item _logged($wrapped_fn, $event, $log_item, $log_item_pre)
+
+Invokes C<log_hook> when it's defined, before and after calling C<$wrapped_fn>
+with no arguments, with the 'pre_' and 'post_' prefixes to the event name.
+
+C<$log_item> can be a fixed string or a function of one argument returning
+the string to be logged. The argument passed into the function is the value
+returned by the C<$wrapped_fn>.
+
+In case there is no C<$log_item_pre> to be called on the 'pre_' event, C<$log_item>
+will be used instead, with no arguments.
+
+For performance reasons, the C<$log_item> and C<$log_item_pre> - when coderefs - aren't
+called; instead they are passed as-is to the C<$log_hook> for lazy evaluation.
+
+=cut
+
+sub _logged {
+    my ($self, $f, $e, $l, $lp) = @_;
+    my $hook = $self->log_hook;
+
+    return $f->() if ! defined $hook;
+
+    $lp //= $l;
+    my $pre = (ref $lp eq 'CODE') ? $lp : _appending_wrap($lp);
+    my $post = (ref $l eq 'CODE') ? $l : _appending_wrap($l);
+    $hook->("pre_$e", $pre);
+    if (wantarray) {
+        my @rv = $f->();
+        $hook->("post_$e", sub { return $l->(\@rv); });
+        return @rv;
+    }
+    else {
+        my $rv = $f->();
+        $hook->("post_$e", sub { return $l->($rv); });
+        return $rv;
+    }
+};
+
 =item _wrap_widget($_id)
 
 Finds all matching widget selectors to wrap the driver element in.
